#!/usr/bin/env python3

from __future__ import annotations
import difflib

from uvpipx.uvpipx_bins import relink_bins

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import json
import re
from typing import Union

from uvpipx import config
from uvpipx.internal_libs.misc import (
    log_info,
    shell_run,
    shell_run_elapse,
)


def upgrade(
    package_name_ref: str,
    *,
    venv_name: Union[None, str] = None,
) -> None:
    package_name = re.search(r"([^=<>]+)(==|>)*", package_name_ref)[1]
    venv_name_ = venv_name or package_name
    pck_venv = config.uvpipx_venvs / venv_name_

    if not (pck_venv / ".venv").exists():
        msg = "{pck_venv} not exist or ready"
        raise RuntimeError(msg)

    with (pck_venv / "uvpipx.json").open() as outfile:
        uvpipx_dict = json.load(
            outfile,
        )
    expose_bin_names_ = uvpipx_dict["bin_names"]
    package_name_ref_ = uvpipx_dict["package_name_ref"]

    rc, stdout, stderr = shell_run(f"cd {pck_venv}; uv pip freeze")
    old_vers = sorted(line for line in stdout.split("\n"))

    shell_run_elapse(
        f"cd {pck_venv}; uv pip install --upgrade {package_name_ref_}",
        f" 📥 uv pip install {package_name_ref_} in uvpipx venv {venv_name_}",
    )
    log_info(f" 🟢 uvpipx venv {venv_name_} with {package_name} ready")
    shell_run(f"cd {pck_venv}; uv pip freeze > requirements.txt")
    rc, stdout, stderr = shell_run(f"cd {pck_venv}; uv pip freeze")
    new_vers = sorted(line for line in stdout.split("\n"))

    diff_vers = [
        n
        for n in difflib.ndiff(old_vers, new_vers)
        if n[:2] in ["+ ", "- "]
    ]

    if diff_vers:
        log_info("\n 🏗️  changes")
        log_info("\n".join(f"   {n}" for n in diff_vers))
    else:
        log_info("\n ⭕ no change")

    log_info("")

    with (pck_venv / "uvpipx.json").open("w") as outfile:
        json.dump(uvpipx_dict, outfile, indent=4, default=str)

    # log_info(" 🎯 Re-Exposing program change")
    # relink_bins(package_name, expose_bin_names=expose_bin_names_, venv_name=venv_name)



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


from uvpipx.internal_libs.args import Arg, ArgParser, ArgParserMode

arg_parser: dict[str, ArgParser] = {}

verbose_arg = Arg(
    "--verbose",
    mode="count",
    help="""Define the output verbosity, you can use --verbose up to three time to increase the verbosity""",
)

help_arg = Arg(
    "--help",
    mode="bool/true/stopper",
    help="""Show help""",
)

arg_parser["install"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python package name to install (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        Arg(
            "--expose",
            mode="array",
            default=["*"],
            help="""By default, all executable program in bin will be expose to `uvpipx_local_bin`\nYou can use --expose jc to expose only this program. you can use many --expose\n--expose _ tell to expose nothing""",
        ),
        Arg(
            "--force",
            mode="bool/true",
            help="""Allow to force reinstall""",
        ),
        verbose_arg,
        help_arg,
    ],
    help="Install a python package",
)

arg_parser["uninstall"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python package name to uninstall (for example "jc")""",
        ),
        verbose_arg,
        help_arg,
    ],
    help="Uninstall a python package",
)

arg_parser["venv"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python package name of the venv you want to use (for example "jc")""",
        ),
        verbose_arg,
        help_arg,
    ],
    help="Run a command in the venv of the python package",
)

arg_parser["ensurepath"] = ArgParser(
    [
        verbose_arg,
        help_arg,
    ],
    help="Configure PATH for uvpipx",
)
arg_parser["list"] = ArgParser(
    [
        verbose_arg,
        help_arg,
    ],
    help="Show the list of python package installed",
)
arg_parser["info"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python package name to get information (for example "jc")""",
        ),
        verbose_arg,
        help_arg,
        Arg("--get-venv", mode="bool/true"),
    ],
    help="Show information of the python package",
)

arg_parser["upgrade"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python package name to upgrade (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        verbose_arg,
        help_arg,
    ],
    help="Upgrade a python package",
)

arg_parser["upgrade-all"] = ArgParser(
    [
        verbose_arg,
        help_arg,
    ],
    help="Upgrade all python package",
)

arg_parser["inject"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python main package name (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        Arg(
            "inject_python_pkg",
            help="""The python package name to inject (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        verbose_arg,
        help_arg,
    ],
    mode=ArgParserMode.AUTO_EXTRA_ARGS,
    help="Inject one or many python package(s) in main python package",
)

arg_parser["uninject"] = ArgParser(
    [
        Arg(
            "python_pkg",
            help="""The python main package name (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        Arg(
            "uninject_python_pkg",
            help="""The python package name to uninject (for example "jc")\nBut you can also give version using "jc==1.25.2" """,
        ),
        verbose_arg,
        help_arg,
    ],
    mode=ArgParserMode.AUTO_EXTRA_ARGS,
    help="Uninject one or many python package(s) in main python package",
)



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.2.0"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
import platform
import shutil
import subprocess  # nosec: B404
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union

from uvpipx.internal_libs.colors import Color, Painter


# Définir le chemin du répertoire à parcourir
def find_executable(dir_path: Path, allow_symlink: bool = False) -> List[str]:
    return [
        file
        for file in dir_path.iterdir()
        if file.is_file()
        and os.access(file, os.X_OK)
        and (not file.is_symlink() or allow_symlink)
    ]


class Elapser:
    """
    A context manager for measuring elapsed time.

    Explanation:
    This class can be used as a context manager to measure the elapsed time within the context block.

    Args:
        None

    Returns:
        None

    Examples:
        with Elapser() as elapser:
            # Code block to measure elapsed time
    """

    def __init__(self) -> None:
        self.start = None
        self.end = None
        self.interval_seconds = None
        self.elapsed_second = None

    def __enter__(self) -> Elapser:
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args: list) -> Union[bool, None]:  # , *args
        self.end = time.perf_counter()
        self.interval_seconds = self.end - self.start
        self.elapsed_second = f"{self.interval_seconds:.3f} seconds"


def log_debug(messages: str, level: int = 1) -> None:
    """
    Logs a message with an optional verbosity level.

    Args:
        message (str): The message to be logged.
        level (int, optional): The verbosity level of the message. Defaults to 1.

    Returns:
        None
    """

    if level <= int(os.getenv("UVPIPX_SHOW_DEBUG_LEVEL", "0")):
        text_info = Painter.color_str(f"DEB{level}", Color.BLUE)
        prefix = (
            f"[{text_info}] "
            if os.getenv("UVPIPX_SHOW_LOG_PREFIX", "0").lower() in ["1", "true"]
            else ""
        )

        for message in messages.split("\n"):
            print(f"{prefix}{message}")


def log_info(messages: str) -> None:
    """
    Logs an informational message.

    Explanation:
    This function prints an informational message with a specific format to the console.

    Args:
        message (str): The message to be logged.

    Returns:
        None
    """
    text_info = Painter.color_str("INFO", Color.BRIGHT_CYAN, Color.ST_BOLD)
    prefix = (
        f"[{text_info}] "
        if os.getenv("UVPIPX_SHOW_LOG_PREFIX", "0").lower() in ["1", "true"]
        else ""
    )

    for message in messages.split("\n"):
        print(f"{prefix}{message}")


def log_warm(messages: str) -> None:
    """
    Logs a warm message.

    Args:
        message (str): The warm message to be logged.

    Returns:
        None
    """
    text_info = Painter.color_str("WARM", Color.BRIGHT_YELLOW, Color.ST_BOLD)
    prefix = (
        f"[{text_info}] "
        if os.getenv("UVPIPX_SHOW_LOG_PREFIX", "0").lower() in ["1", "true"]
        else ""
    )

    for message in messages.split("\n"):
        print(f"{prefix}{message}")


def log_error(messages: str) -> None:
    """
    Logs a error message.

    Args:
        message (str): The error message to be logged.

    Returns:
        None
    """
    text_info = Painter.color_str("ERR ", Color.BRIGHT_RED, Color.ST_BOLD)
    prefix = (
        f"[{text_info}] "
        if os.getenv("UVPIPX_SHOW_LOG_PREFIX", "0").lower() in ["1", "true"]
        else ""
    )

    for message in messages.split("\n"):
        print(f"{prefix}{message}")


def shell_run(
    command: str,
    *,
    env: Union[None, Dict[str, str]] = None,
    raise_on_error: bool = True,
) -> Tuple[int, Union[bytes, str], Union[bytes, str]]:
    """
    Executes a shell command and returns the result.

    Explanation:
    This function runs a shell command and captures the standard output, standard error, and return code.

    Args:
        command (List[str]): The shell command to be executed.
        raise_on_error (bool): Flag to raise an error if the command execution fails. Default is False.

    Returns:
        Tuple[int, str, str]: A tuple containing the return code, standard output, and standard error.
    """
    env_ = cmd_prepare_env(env)
    encoding = cmd_prepare_encoding()

    with subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,  # noqa: S602
        env=env_,
    ) as proc:  # nosec: B602
        try:
            stdout, stderr = proc.communicate()
        except subprocess.SubprocessError:
            if raise_on_error:
                raise

        rc = proc.returncode
        stdout = stdout.decode(encoding)
        stderr = stderr.decode(encoding)
        if rc != 0 and raise_on_error:
            short_msg = f"{stderr:2000}".rstrip()
            raise RuntimeError(
                f"🔴 Command failed with return code {rc} {short_msg}..."
            )

    return rc, stdout, stderr


def cmd_run(
    cwd: str,
    command: str,
    *,
    env: Union[None, Dict[str, str]] = None,
    raise_on_error: bool = True,
    raw_pipe: bool = False,
) -> Tuple[int, Union[bytes, str], Union[bytes, str]]:
    """
    Executes a shell command and returns the result.

    Explanation:
    This function runs a shell command and captures the standard output, standard error, and return code.

    Args:
        command (List[str]): The shell command to be executed.
        raise_on_error (bool): Flag to raise an error if the command execution fails. Default is False.

    Returns:
        Tuple[int, str, str]: A tuple containing the return code, standard output, and standard error.
    """
    env_ = cmd_prepare_env(env)
    encoding = cmd_prepare_encoding()
    pipe_type = None if raw_pipe else subprocess.PIPE

    with subprocess.Popen(
        command,  # noqa: S603
        stdout=pipe_type,
        stderr=pipe_type,
        cwd=cwd,
        env=env_,
    ) as proc:  # nosec: B603
        try:
            if raw_pipe:
                stdout, stderr = (None, None)
                proc.wait()
            else:
                stdout, stderr = proc.communicate()
                stdout = stdout.decode(encoding)
                stderr = stderr.decode(encoding)
        except subprocess.SubprocessError:
            if raise_on_error:
                raise

    rc = proc.returncode
    if rc != 0 and raise_on_error:
        short_msg = f"{stderr:2000}".rstrip()
        raise RuntimeError(f"🔴 Command failed with return code {rc} {short_msg}...")

    return rc, stdout, stderr


def cmd_prepare_env(env: Dict[str, str]) -> Dict[str, str]:
    env_ = env
    if env_ is None:
        env_ = os.environ.copy()
        k_to_del = [k for k in env_ if "VIRTUAL_ENV" in k]
        for k in k_to_del:
            del env_[k]
    return env_


def cmd_prepare_encoding() -> str:
    os_type = platform.system()
    encoding = "utf-8"
    if os_type == "Windows":
        encoding = "utf-16"
    #     if "powershell" in command[0]:
    #         # command = ["powershell", "-Command"] + command
    #         encoding = 'utf-16'
    #     else:
    #         # command = ["cmd", "/c"] + command
    #         encoding = 'oem'
    return encoding


def shell_run_elapse(
    command: str,
    message: str,
    *,
    raise_on_error: bool = True,
) -> None:
    with Elapser() as ela:
        rc, std_o, std_e = shell_run(command, raise_on_error=raise_on_error)

    log_info(f"{message}   ⏱️  {ela.elapsed_second}")


def command_exists(cmd: str) -> bool:
    return shutil.which(cmd) is not None



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import List, Union

RE_PIP_REQ = re.compile(
    r"""
    (?P<name>[A-Za-z0-9._-]]+)\s*             # Nom du projet
    (\[(?P<extras>[A-Za-z0-9._- ,]+)\])?\s*     # Extras optionnels, entourés de []
    (?P<version_specifiers>[^;]+)?\s*             # Spécifications des versions
    (;\s*(?P<environment_marker>.+))?       # Environnement optionnel, après ;
""",
    re.VERBOSE,
)


@dataclass
class Requirement:
    # PEP 508 – Dependency specification for Python Software Packages
    name: str
    extras: Union[None, List[str]] = None
    version_specifiers: Union[None, List[str]] = None
    environment_marker: Union[None, str] = None

    @staticmethod
    def split_str(data: str) -> List[str]:
        return [element.strip() for element in data.split(",")] if data else []

    @classmethod
    def from_line(cls, line: str) -> "Requirement":
        match = RE_PIP_REQ.match(line)
        if not match:
            raise RuntimeError(f"Line {line} not match PIP_REQ")

        return cls(
            name=match.group("name"),
            extras=Requirement.split_str(match.group("extras")),
            specs=Requirement.split_str(match.group("specs")),
            marker=match.group("marker"),
        )


# requirements = [
#     "SomeProject",
#     "SomeProject == 1.3",
#     "SomeProject >= 1.2, < 2.0",
#     "SomeProject [foo, bar]",
#     "SomeProject ~= 1.4.2",
#     "SomeProject == 5.4 ; python_version < '3.8'",
#     "SomeProject ; sys_platform == 'win32'",
#     'requests [security] >= 2.8.1, == 2.8.* ; python_version < "2.7"',
#     "SomeProject==1.3",
#     "SomeProject>= 1.2, < 2.0",
#     "SomeProject[foo, bar]",
#     "SomeProject~=1.4.2",
#     "SomeProject==5.4;python_version<'3.8'",
#     "SomeProject;sys_platform=='win32'",
#     'requests[security]>=2.8.1,==2.8.*;python_version<"2.7"',
# ]



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
from pathlib import Path
from typing import Union

# $XDG_CACHE_HOME = $HOME/.cache
# PIPX_HOME= so default to /home/gaetan/.local/pipx
# PIPX_LOCAL_VENVS=/home/gaetan/.local/pipx/venvs
# PIPX_VENV_CACHEDIR=/home/gaetan/.local/pipx/.cache
# PIPX_DEFAULT_PYTHON=/usr/bin/python3
# USE_EMOJI=true


def env_to_path(env_name: str, default: Union[str, None] = None) -> Path:
    return Path(os.path.expandvars(os.environ.get(env_name, default)))


ci_project_dir = env_to_path("CI_PROJECT_DIR", ".")
home = env_to_path("HOME")
uvpipx_home = env_to_path("UVPIPX_HOME", home / ".local/uv-pipx")
uvpipx_venvs = env_to_path("UVPIPX_LOCAL_VENVS", uvpipx_home / "venvs")

default_local_bin = home / ".local/bin"
if home == Path("/root"):
    default_local_bin = "/usr/local/bin"

uvpipx_local_bin = env_to_path("UVPIPX_BIN_DIR", default_local_bin)



import sys

from uvpipx.internal_libs.colors import Painter
from uvpipx.version import show_version

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"

import uvpipx.cmd_launcher as cmd_launcher

# import (
#     ensurepath,
#     info,
#     install,
#     uninstall,
#     uvpipx_list,
#     venv,
#     upgrade, upgrade_all,
#     inject,
# )
from uvpipx.internal_libs.print import max_string_length_per_column, wrap_text_in_table
from uvpipx.uvpipx_args import arg_parser

cmd_map = {"list": "uvpipx_list"}


def show_main_help() -> None:
    print(
        Painter.parse_color_tags("<ST_UNDERLINE>uvpipx general help</ST_UNDERLINE>\n"),
    )

    print("uvpipx cmd [--help] ... parameters ...\n\ncmd can be:")
    help_ = [[f"{k}", f"{v.help}"] for k, v in arg_parser.items()]
    size_col = max_string_length_per_column(help_)

    wrapped = wrap_text_in_table(help_, [size_col[0], 100 - size_col[0]])
    for row in wrapped:
        for ss in [list(sub_row) for sub_row in zip(*row)]:
            print("  " + (" | ".join(ss)).rstrip())

    print()


def main() -> None:
    """uv pipx

    like pipx but with uv ... intent to be miniamist, small and so fast !
    """

    if len(sys.argv) <= 1 or (len(sys.argv) == 2 and sys.argv[1] in ["-h", "--help"]):
        show_version()
        show_main_help()
        sys.exit(0)

    main_cmd = sys.argv[1]
    if main_cmd in arg_parser:
        main_cmd_map = main_cmd.replace("-", "_")
        cmd_function = getattr(cmd_launcher, cmd_map.get(main_cmd_map, main_cmd_map))
        cmd_function(arg_parser[sys.argv[1]])

    else:
        print(f"Unknow command {sys.argv[1]}, below the help")
        show_main_help()

        msg = f"Unknow command {sys.argv[1]}"
        raise RuntimeError(msg)


if __name__ == "__main__":
    main()



__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"

import os
import sys

from uvpipx import (
    uvpipx_all,
    uvpipx_infos,
    uvpipx_inject,
    uvpipx_install,
    uvpipx_upgrade,
)
from uvpipx.internal_libs.args import ArgParser
from uvpipx.internal_libs.misc import Elapser, log_debug, log_info


def common_args(argp: ArgParser) -> None:
    argp.parse(sys.argv[2:])
    os.environ["UVPIPX_SHOW_DEBUG_LEVEL"] = str(
        argp.args["--verbose"].defaulted_value(),
    )
    log_debug(f"Received args {sys.argv}")

    if argp.args["--help"].defaulted_value():
        argp.print_help()
        sys.exit(0)


def install(argp: ArgParser) -> None:
    """install package locally in their own venv"""

    common_args(argp)

    with Elapser() as ela:
        uvpipx_install.install(
            argp.args["python_pkg"].value,
            expose_bin_names=argp.args["--expose"].defaulted_value(),
            force_reinstall=argp.args["--force"].defaulted_value(),
        )

    log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")


def ensurepath(argp: ArgParser) -> None:
    """help to define PATH"""

    common_args(argp)

    uvpipx_infos.ensurepath()


def uvpipx_list(argp: ArgParser) -> None:
    """show the list of uvpipx venv"""

    common_args(argp)

    uvpipx_infos.uvpipx_list()


def info(
    argp: ArgParser,
) -> None:
    """show the list of uvpipx venv"""

    common_args(argp)

    uvpipx_infos.info(
        argp.args["python_pkg"].value,
        get_venv=argp.args["--get-venv"].defaulted_value(),
    )


def uninstall(argp: ArgParser) -> None:
    """uninstall package and their venv"""

    common_args(argp)

    with Elapser() as ela:
        uvpipx_install.uninstall(
            argp.args["python_pkg"].value,
        )

    log_info(f"\n 🏁 Finish uninstall  ⏱️  {ela.elapsed_second}")


def venv(argp: ArgParser) -> None:
    """run command of a package without installing it"""
    common_args(argp)

    uvpipx_install.run_venv_bin(argp.args["python_pkg"].value, argp.extra_args)


def upgrade(argp: ArgParser) -> None:
    """upgrade package locally in their own venv"""

    common_args(argp)

    with Elapser() as ela:
        uvpipx_upgrade.upgrade(
            argp.args["python_pkg"].value,
        )

    log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")


def upgrade_all(argp: ArgParser) -> None:
    """upgrade all package locally in their own venv"""

    common_args(argp)

    with Elapser() as ela:
        uvpipx_all.upgrade_all()

    log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")


def inject(argp: ArgParser) -> None:
    """inject package locally in venv"""

    common_args(argp)

    with Elapser() as ela:
        inject_pkg = [argp.args["inject_python_pkg"].value] + argp.extra_args
        uvpipx_inject.inject(argp.args["python_pkg"].value, inject_pkg)

    log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")


def uninject(argp: ArgParser) -> None:
    """inject package locally in venv"""

    common_args(argp)

    with Elapser() as ela:
        inject_pkg = [argp.args["uninject_python_pkg"].value] + argp.extra_args
        uvpipx_inject.uninject(argp.args["python_pkg"].value, inject_pkg)

    log_info(f"\n 🏁 Finish install  ⏱️  {ela.elapsed_second}")



# import urllib.request


# def download_file(
#     url,
#     destination_filename,
#     cafile=None,
# ) -> None:
#     if url.startswith("file:/"):
#         msg = "Download from file is not allowed !"
#         raise RuntimeError(msg)

#     with urllib.request.urlopen(url, cafile=cafile) as response, destination_filename.open(  # nosec: B310
#          "wb",
#     ) as file:
#         data = response.read(64 * 1024)
#         while data:
#             file.write(data)
#             data = response.read(64 * 1024)
#         file.write(data)


# def download_as_bytes(
#     url,
#     destination_filename,
#     cafile=None,
# ):
#     if url.startswith("file:/"):
#         msg = "Download from file is not allowed !"
#         raise RuntimeError(msg)

#     with urllib.request.urlopen(url, cafile=cafile) as response:  # nosec: B310
#         data = response.read(64 * 1024)
#         while data:
#             data += response.read(64 * 1024)

#     return data



from __future__ import annotations

from typing import List, Optional, Union

from uvpipx.internal_libs.misc import (
    log_error,
    log_info,
    log_warm,
)
from uvpipx.UvPipxEnv import ExposeBin, UvPipx


def relink_bins(
    package_name: str,
    *,
    expose_bin_names: Optional[List[str]] = None,
    venv_name: Union[None, str] = None,
) -> None:
    expose_bin_names_ = expose_bin_names or ["*"]

    uvp = UvPipx(
        package_name=package_name,
        venv_name=venv_name,
        # expose_bin_names=expose_bin_names_,
    )
    new_bins = uvp.dir_bins_to_expose(expose_bin_names_)

    exposed_bins_state: dict[str, list[ExposeBin]] = {
        "new": [],
        "exist_me": [],
        "exist_not_me": [],
        "not_exist": [],
        "del": [],
    }

    for new_bin in new_bins:
        if new_bin.link_exists():
            if new_bin.valid():
                exposed_bins_state["exist_me"].append(new_bin)
            else:
                exposed_bins_state["exist_not_me"].append(new_bin)
        else:
            if new_bin.exists():
                exposed_bins_state["new"].append(new_bin)
            else:
                exposed_bins_state["not_exist"].append(new_bin)

    new_bin_names = set(n.venv_bin_name for n in new_bins)
    for old_bin in uvp.saved_bin_exposed():
        if old_bin.venv_bin_name not in new_bin_names:
            exposed_bins_state["del"].append(old_bin)

    uvpipx_new_bins = apply_relink_bins(exposed_bins_state)
    uvp.uvpipx_dict["exposed_bins"] = sorted(uvpipx_new_bins)

    uvp.save(uvp.uvpipx_dict)


def apply_relink_bins(exposed_bins_state: dict[str, list[ExposeBin]]):  # -> list:
    uvpipx_new_bins = []
    for exposing in exposed_bins_state["new"]:
        log_info(
            f"  📁 Exposing program {exposing.venv_bin_path()} to {exposing.local_bin_path}"
        )
        exposing.link()
        uvpipx_new_bins.append((str(exposing.venv_bin_path()), exposing.local_bin_path))

    for exposing in exposed_bins_state["del"]:
        log_info(
            f"  🗑️  Remove exposed program {exposing.local_bin_path} -> {exposing.venv_bin_path()}"
        )
        exposing.unlink()

    for exposing in exposed_bins_state["exist_me"]:
        # log_info(f"Exposing program {exposing.venv_bin_name} to {exposing.local_bin_path}")
        uvpipx_new_bins.append((str(exposing.venv_bin_path()), exposing.local_bin_path))

    for exposing in exposed_bins_state["exist_not_me"]:
        log_warm(
            f"  🟡 Link {exposing.local_bin_path} already exist. Skipping {exposing.venv_bin_name}!"
        )

    for exposing in exposed_bins_state["not_exist"]:
        log_error(f"  🔴 {exposing.venv_bin_name} not exist. Cannot expose !")

    return uvpipx_new_bins



from __future__ import annotations

from uvpipx.version import show_version

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import json
import os
from pathlib import Path
from typing import Union

from uvpipx import config
from uvpipx.internal_libs.misc import log_info, shell_run


def ensurepath() -> None:
    path_set = os.environ["PATH"].split(":")
    if str(config.uvpipx_local_bin) + "x" in path_set:
        log_info("🟢 Configuration of PATH already OK")
        return

    profile_path = Path(os.environ["HOME"]) / ".profile"

    profile_content = Path(profile_path).read_text()
    if "# Added by uvpipx" in profile_content:
        log_info(f"""⚠️  Configuration already in {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")
        return

    line_to_add = ""
    if Path(profile_path).read_text()[-1] != "\n":
        line_to_add = "\n"

    line_to_add += f"""# Added by uvpipx
export PATH=$PATH:{config.uvpipx_local_bin}\n"""

    with profile_path.open("a") as file:
        file.write(line_to_add)

    log_info(f"""Configuration added to {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")


def _info(pck_venv: Path) -> str:
    def get_version(pck_name, stdout_line):
        vers = next(
            (line for line in stdout_line if pck_name in line),
            None,
        )
        if vers is None:
            vers = f"{pck_name} Unknow version"

        return vers

    uvpipx_dict = {}
    with (pck_venv / "uvpipx.json").open() as intfile:
        uvpipx_dict = json.load(
            intfile,
        )
    rc, stdout, stderr = shell_run(f"cd {pck_venv}; uv pip freeze")
    stdout_line =[line for line in stdout.split("\n")]
    main_vers = get_version(uvpipx_dict["package_name"], stdout_line)
        # vers = "unable to retreive package information"

    injected_vers=[]
    for pkg_injected in uvpipx_dict.get("injected_package",[]):
        injected_vers.append(get_version(pkg_injected, stdout_line))
    injected_vers = sorted(injected_vers)

    bins = "\n".join(
        f"   ✅ {bin_.split('/')[-1]}" for _, bin_ in uvpipx_dict["exposed_bins"]
    )
    if not bins:
        bins = "   ❌ Nothing exposed"

    output = f""" 📦 {uvpipx_dict['package_name']} ({main_vers}) in venv {uvpipx_dict['uvpipx_package_path']}"""
    if injected_vers:
        output_inject = "\n".join(f"""   📦 {pkg_ver}""" for pkg_ver in injected_vers)
        output += f"""

 🎯 Injected packages
{output_inject}"""
        
    output += f"""

 🎯 Exposed program
{bins}\n"""
    
    return output



def info(
    package_name: str,
    *,
    venv_name: Union[None, str] = None,
    get_venv: bool = False,
) -> None:
    venv_name_ = venv_name or package_name
    pck_venv = config.uvpipx_venvs / venv_name_

    info = str(pck_venv / ".venv") if get_venv else _info(pck_venv)

    log_info(info)


def uvpipx_list() -> None:
    infos = f"""📍 uvpipx venvs are in {config.uvpipx_venvs}
💻 apps are exposed at {config.uvpipx_local_bin} 

"""
    show_version()

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            infos += _info(pck_venv)
            infos += "\n\n"
            nb += 1

    if nb == 0:
        infos += "⭕ Nothing is installed !"

    log_info(infos)



#!/usr/bin/env python3

from __future__ import annotations

from uvpipx.uvpipx_bins import relink_bins

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import json
import re
import shutil
import sys
from pathlib import Path
from typing import List, Optional, Union

from uvpipx import config
from uvpipx.internal_libs.misc import (
    cmd_prepare_env,
    cmd_run,
    log_info,
    shell_run,
    shell_run_elapse,
)


def install(
    package_name_ref: str,
    *,
    expose_bin_names: Optional[List[str]] = None,
    venv_name: Union[None, str] = None,
    force_reinstall: bool = False,
) -> None:
    package_name = re.search(r"([^=<>]+)(==|>)*", package_name_ref)[1]
    venv_name_ = venv_name or package_name
    expose_bin_names_ = expose_bin_names

    pck_venv = config.uvpipx_venvs / venv_name_

    if expose_bin_names_ is None:
        expose_bin_names_ = ["*"]

    if not force_reinstall and (pck_venv / "uvpipx.json").exists():
        log_info(
            f"⚠️  {package_name_ref} already installed. need to use option --force to reinstall"
        )
        return

    uvpipx_dict = {
        "package_name_ref": package_name_ref,
        "package_name": package_name,
        "venv_name": venv_name,
        "bin_names": expose_bin_names_,
    }

    # if not command_exists("uv"):
    #     log_info("Install uv")
    #     shell_run_elapse("pip install uv", "uv installed")

    (config.uvpipx_venvs / venv_name_).mkdir(exist_ok=True, parents=True)
    uvpipx_dict["uvpipx_package_path"] = pck_venv

    if not (pck_venv / ".venv").exists():
        shell_run_elapse(
            f"uv venv {pck_venv / '.venv'}",
            f" 📦 uv venv {pck_venv} created",
        )

    shell_run_elapse(
        f"cd {pck_venv}; uv pip install {package_name_ref}",
        f" 📥 uv pip install {package_name_ref} in uvpipx venv {venv_name_}",
    )
    log_info(f" 🟢 uvpipx venv {venv_name_} with {package_name} ready")
    shell_run(f"cd {pck_venv}; uv pip freeze > requirements.txt")
    log_info("")

    with (pck_venv / "uvpipx.json").open("w") as outfile:
        json.dump(uvpipx_dict, outfile, indent=4, default=str)

    log_info(" 🎯 Exposing program")
    relink_bins(package_name, expose_bin_names=expose_bin_names_, venv_name=venv_name)


# def reinstall(
#     package_name_ref: str,
#     *,
#     expose_bin_names: Optional[List[str]] = None,
#     venv_name: Union[None, str] = None,
# ) -> None:
#     return install(
#         package_name_ref,
#         expose_bin_names=expose_bin_names,
#         venv_name=venv_name,
#         force_reinstall=True,
#     )


def run_venv_bin(
    package_name: str,
    cmdline: str,
    *,
    venv_name: Union[None, str] = None,
) -> None:
    """venv is specific to uvpipx. it can replace inject, runpip (oups runuv), uninject

    in fact runpip should be replace by uvpipx venv "truc" uv -- pip install me
    """
    venv_name_ = venv_name or package_name
    pck_venv = config.uvpipx_venvs / venv_name_

    if not (pck_venv / ".venv").exists():
        msg = "{pck_venv} not exist or ready"
        raise RuntimeError(msg)

    env = cmd_prepare_env(None)
    env["VIRTUAL_ENV"] = str(pck_venv / ".venv")
    env["PATH"] = str(pck_venv / ".venv" / "bin") + ":" + env["PATH"]
    if "PYTHONHOME" in env:
        del env["PYTHONHOME"]

    rc, stdo, stde = cmd_run(
        # pck_venv / ".venv" / "bin",
        Path.cwd(),
        cmdline,
        env=env,
        raw_pipe=True,
    )
    # print(stdo)
    # print(stde, file=sys.stderr)
    sys.exit(rc)


def uninstall(package_name: str, *, venv_name: Union[None, str] = None) -> None:
    venv_name_ = venv_name or package_name
    pck_venv = config.uvpipx_venvs / venv_name_

    if not (pck_venv / ".venv").exists():
        msg = f"🔴 {package_name} not exist (path {pck_venv})"
        raise RuntimeError(msg)

    with (pck_venv / "uvpipx.json").open() as outfile:
        uvpipx_dict = json.load(
            outfile,
        )

    log_info(f"🗑️  Remove exposed program")
    relink_bins(package_name, expose_bin_names=["_"], venv_name=venv_name)

    log_info(f"  🗑️  Remove uvpipx venv {pck_venv}")
    shutil.rmtree(pck_venv)



from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.2.0"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"

import textwrap
from typing import List


class NewLinePreservingWrapper(textwrap.TextWrapper):
    def wrap(self, text: str) -> List[str]:
        # Split the original text by new lines, process each part, then combine
        wrapped_lines = []
        for part in text.split("\n"):
            wrapped_lines.extend(super().wrap(part))
        return wrapped_lines

    def fill(self, text: str) -> str:
        # Use the wrap method and join the results with new lines
        return "\n".join(self.wrap(text))


def max_string_length_per_column(table: List[List[str]]) -> List[int]:
    """
    Calculates the maximum string length for each column in a 2D table.

    Args:
        table (list of list of str): A 2D list where each sublist represents a row of the table.

    Returns:
        list of int: A list containing the maximum string length found in each column.

    Raises:
        ValueError: If the table is empty or rows have inconsistent numbers of columns.
    """

    num_columns = len(table[0])

    # Initialize a list to store the maximum length found in each column.
    max_lengths = [0] * num_columns

    # Compute the maximum length for each column.
    for col_index in range(num_columns):
        # Using max() with a generator expression to find the max length in each column.
        max_lengths[col_index] = max(len(row[col_index]) for row in table)

    return max_lengths


def wrap_text_in_table(table: str, column_widths: int) -> List[str]:
    """
    Wraps text in each cell of a table to the specified widths for each column and ensures each line in a cell
    is padded with spaces to maintain the column width. Also ensures all cells in a row have the same number of lines.

    Parameters:
        table (list of list of str): The table to wrap text in, with each cell as a string.
        column_widths (list of int): The maximum width for each column.

    Returns:
        list of list of list of str: The table with text wrapped and aligned in each cell.
    """
    # Create a TextWrapper object for each column
    wrappers = [NewLinePreservingWrapper(width=width) for width in column_widths]

    # Process each row in the table
    wrapped_table = []
    for row in table:
        wrapped_row = []
        max_lines_per_row = 0

        # Wrap text in each cell and determine the max number of lines in this row
        for content, wrapper in zip(row, wrappers):
            wrapped_text = wrapper.wrap(text=content)
            # Pad each line to ensure it has the exact column width
            wrapped_text = [line.ljust(wrapper.width) for line in wrapped_text]
            max_lines_per_row = max(max_lines_per_row, len(wrapped_text))
            wrapped_row.append(wrapped_text)

        # Ensure all cells in this row have the same number of lines
        for i, cell in enumerate(wrapped_row):
            cell.extend(["".ljust(column_widths[i])] * (max_lines_per_row - len(cell)))

        wrapped_table.append(wrapped_row)

    return wrapped_table



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Union

from uvpipx import config
from uvpipx.internal_libs.misc import find_executable

# cat /home/gaetan/.local/uv-pipx/venvs/jc/uvpipx.json


@dataclass
class ExposeBin:
    venv_path: Path
    venv_bin_name: str
    local_bin_path: Union[None, Path] = None

    def venv_bin_path(self) -> Path:
        return self.venv_path / ".venv/bin" / self.venv_bin_name

    def exists(self) -> bool:
        return self.venv_bin_path().exists()

    def link_exists(self) -> bool:
        if self.local_bin_path is None:
            msg = "Link not defined"
            raise RuntimeError(msg)

        return self.local_bin_path.exists()

    def valid(self) -> bool:
        if self.local_bin_path.exists():
            return self.local_bin_path.is_symlink() and (
                self.venv_bin_path() == self.local_bin_path.resolve()
            )

        return False

    @classmethod
    def build_local_bin_path(
        cls,
        venv_path: Path,
        venv_bin_name: str,
        alias_bin_name: Union[None, str] = None,
    ) -> Path:
        bin_name = alias_bin_name or venv_bin_name
        return cls(venv_path, venv_bin_name, Path(config.uvpipx_local_bin) / bin_name)

    def link(self) -> None:
        os.symlink(self.venv_bin_path(), self.local_bin_path)

    def unlink(self, if_valid: bool = True) -> None:
        if (if_valid and self.valid()) or (not if_valid and self.exists()):
            self.local_bin_path.unlink()


@dataclass
class UvPipx:
    package_name: str
    venv_name: Union[None, str]
    exposed_bins: Union[List[ExposeBin]] = None

    def __post_init__(self):
        self.venv_name_ = self.venv_name or self.package_name
        self.pck_venv = config.uvpipx_venvs / self.venv_name_
        if self.exists():
            self.load()

    def exists(self) -> bool:
        return self.pck_venv.exists()

    def saved_bin_exposed(self) -> List[ExposeBin]:
        return [
            ExposeBin(
                venv_path=self.pck_venv,
                venv_bin_name=Path(exposed_bin[0]).name,
                local_bin_path=Path(exposed_bin[1]),
            )
            for exposed_bin in self.uvpipx_dict.get("exposed_bins", [])
        ]

    def load(self) -> None:
        # expose_bin_names_ = expose_bin_names or ["*"]

        with (self.pck_venv / "uvpipx.json").open() as outfile:
            self.uvpipx_dict = json.load(outfile)

    def save(self, uvpipx_dict: dict[str, Any]) -> None:
        with (self.pck_venv / "uvpipx.json").open("w") as outfile:
            json.dump(uvpipx_dict, outfile, indent=4, default=str)

    def dir_bins_to_expose(self, expose_bin_names: list[str]) -> list[ExposeBin]:
        expose_bin_names_ = expose_bin_names
        if len(expose_bin_names_) == 1:
            if expose_bin_names_[0] == "*":
                exe = find_executable(self.pck_venv / ".venv" / "bin")
                expose_bin_names_ = [
                    s.name for s in exe if not s.name.startswith("python")
                ]
            elif expose_bin_names_[0] == "_":
                expose_bin_names_ = []

        prep_expose_bins: list[ExposeBin] = []
        for bin_name in expose_bin_names_:
            venv_bin_name, local_bin_name = (
                bin_name.split(":") if ":" in bin_name else (bin_name, bin_name)
            )
            prep_expose_bins.append(
                ExposeBin.build_local_bin_path(
                    self.pck_venv, venv_bin_name, local_bin_name
                )
            )

        return prep_expose_bins
        # self.exposed_bins = prep_expose_bins



#!/usr/bin/env python3

from __future__ import annotations

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import json
import re
from typing import Union

from uvpipx import config
from uvpipx.internal_libs.misc import (
    log_info,
    shell_run,
    shell_run_elapse,
)


def inject(
    package_main_name: str,
    lst_package_name_ref: list[str],
    *,
    # expose_bin_names: Optional[List[str]] = None,
    venv_name: Union[None, str] = None,
) -> None:
    package_name = package_main_name
    # re.search(r"([^=<>]+)(==|>)*", lst_package_name_ref)[1]
    venv_name_ = venv_name or package_main_name
    pck_venv = config.uvpipx_venvs / venv_name_

    # expose_bin_names_ = expose_bin_names
    # if expose_bin_names_ is None:
    #     expose_bin_names_ = ["*"]

    if not (pck_venv / ".venv").exists():
        msg = "{pck_venv} not exist or ready"
        raise RuntimeError(msg)

    with (pck_venv / "uvpipx.json").open() as outfile:
        uvpipx_dict = json.load(
            outfile,
        )

    injected_package = {
        re.search(r"([^=<>]+)(==|>)*", pck_name)[1]: pck_name
        for pck_name in lst_package_name_ref
    }
    pre_injected_package = uvpipx_dict.get("injected_package", {})

    for pck_name in injected_package:
        if pck_name in pre_injected_package:
            msg = f"🔴 {pck_name} already injected"
            raise RuntimeError(msg)

    pip_packages_spec = " ".join(lst_package_name_ref)
    shell_run_elapse(
        f"cd {pck_venv}; uv pip install --upgrade {pip_packages_spec}",
        f" 📥 uv pip install {pip_packages_spec} in uvpipx venv {venv_name_}",
    )
    log_info(f" 🟢 injected {lst_package_name_ref}")
    shell_run(f"cd {pck_venv}; uv pip freeze > requirements.txt")
    log_info("")

    tmp_dict = {**pre_injected_package, **injected_package}
    uvpipx_dict["injected_package"] = {k: tmp_dict[k] for k in sorted(tmp_dict)}
    with (pck_venv / "uvpipx.json").open("w") as outfile:
        json.dump(uvpipx_dict, outfile, indent=4, default=str)

    # log_info(" 🎯 Re-Exposing program change")
    # relink_bins(package_name, expose_bin_names=expose_bin_names_, venv_name=venv_name)


def uninject(
    package_main_name: str,
    lst_package_name_ref: list[str],
    *,
    venv_name: Union[None, str] = None,
) -> None:
    venv_name_ = venv_name or package_main_name
    pck_venv = config.uvpipx_venvs / venv_name_

    if not (pck_venv / ".venv").exists():
        msg = "{pck_venv} not exist or ready"
        raise RuntimeError(msg)

    with (pck_venv / "uvpipx.json").open() as outfile:
        uvpipx_dict = json.load(
            outfile,
        )

    pre_injected_package = uvpipx_dict.get("injected_package", {})

    uninjected_package = {
        re.search(r"([^=<>]+)(==|>)*", pck_name)[1]: pck_name
        for pck_name in lst_package_name_ref
    }

    for pck_name in uninjected_package:
        if pck_name not in pre_injected_package:
            msg = f"🔴 {pck_name} is not injected"
            raise RuntimeError(msg)

    futur_injected_package = {
        pck_name: pck_ref
        for pck_name, pck_ref in pre_injected_package.items()
        if pck_name not in uninjected_package
    }

    pip_packages_spec = " ".join(uninjected_package.keys())
    shell_run_elapse(
        f"cd {pck_venv}; uv pip uninstall {pip_packages_spec}",
        f" 📥 uv pip iunnstall {pip_packages_spec} in uvpipx venv {venv_name_}",
    )
    log_info(f" 🟢 iunnjected {lst_package_name_ref}")
    shell_run(f"cd {pck_venv}; uv pip freeze > requirements.txt")
    log_info("")

    uvpipx_dict["injected_package"] = sorted(futur_injected_package)

    with (pck_venv / "uvpipx.json").open("w") as outfile:
        json.dump(uvpipx_dict, outfile, indent=4, default=str)

    # log_info(" 🎯 Re-Exposing program change")
    # relink_bins(package_name, expose_bin_names=expose_bin_names_, venv_name=venv_name)



"""intend to be independent(internal one file) args parser ... be cool enougth"""

from __future__ import annotations

from uvpipx.internal_libs.colors import Painter

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.2.0"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from uvpipx.internal_libs.print import max_string_length_per_column, wrap_text_in_table


@dataclass
class Arg:
    name: str
    mode: Union[None, str] = None
    type_: Any = str
    default: Any = None
    help: str = ""
    alternate_name: Union[List[str], None] = field(init=False, default_factory=list)
    arg_pos: int = field(init=False, default=-1)
    value: Any = field(init=False, default=None)

    def defaulted_value(self) -> Union[str, None, bool, int]:
        default_ = self.default
        if self.mode and self.mode.startswith("bool/"):
            default_ = False if self.mode.startswith("bool/true") else False

        if self.mode and self.mode.startswith("count"):
            default_ = 0

        return self.value if self.value is not None else default_


class ArgParserMode(Enum):
    STRICT = "STRICT"
    ALLOW_MISSING_POSITIONAL = "ALLOW_MISSING_POSITIONAL"
    AUTO_EXTRA_ARGS = "AUTO_EXTRA_ARGS"


@dataclass
class ArgParser:
    args_def: List[Arg]
    mode: ArgParserMode = ArgParserMode.STRICT
    args_pos_def: Dict[str, Arg] = field(init=False, default_factory=dict)
    args: Dict[str, Arg] = field(init=False, default_factory=dict)
    args_pos: Dict[str, Arg] = field(init=False, default_factory=dict)
    extra_args: List[str] = field(init=False, default_factory=list)
    help: Union[str, None] = None
    _nb_expected_args: int = field(init=False, default=0)

    def __post_init__(self) -> None:
        for arg_def in self.args_def:
            self._process_arg_definition(arg_def)

    def _process_arg_definition(self, arg_def: List[Arg]) -> None:
        self.args[arg_def.name] = arg_def
        if arg_def.name.startswith("--"):
            if not arg_def.alternate_name and arg_def.name[1:3] not in self.args:
                arg_def.alternate_name = [arg_def.name[1:3]]
                self.args[arg_def.alternate_name[0]] = arg_def
            # TODO manage provider alternate_name
        else:
            arg_def.arg_pos = len(self.args_pos_def)
            self.args_pos_def[arg_def.name] = arg_def
            self._nb_expected_args += 1

    def parse(self, received_args: List[str]) -> None:
        i_token = 0

        # print("received_args",received_args)
        stopper = None

        while i_token < len(received_args):
            try:
                nb_consumed_element, arg = self._parse_arg(i_token, received_args)
                if nb_consumed_element == -1:
                    if arg is None:
                        break
                    else:
                        stopper = arg
                        break

                i_token += nb_consumed_element
                if (
                    self.mode == ArgParserMode.AUTO_EXTRA_ARGS
                    and len(self.args_pos) == self._nb_expected_args
                ):
                    break
            except IndexError as e:
                msg = f"Unable to parse arg {received_args[i_token]}"
                raise RuntimeError(
                    msg,
                ) from e

        # if nb_consumed_element == -1 and self.mode in [ArgParserMode.ALLOW_MISSING_POSITIONAL, ArgParserMode.STRICT]:
        if (
            self.mode == ArgParserMode.STRICT
            and stopper is None
            and len(self.args_pos) != self._nb_expected_args
        ):
            msg = "Missing some parameter"
            raise RuntimeError(
                msg,
            )  # TODO better message indicating the missing list

        if self.mode != ArgParserMode.AUTO_EXTRA_ARGS:
            # if we are here this is because there -- to allow extra extra so we skip one to bypass --
            i_token += 1
        self.extra_args = received_args[i_token:]

        # print("extra_args", self.extra_args)

    def _parse_arg(
        self,
        i_token: int,
        received_args: List[str],
    ) -> Union[Tuple[int, Dict[str, Arg]], Tuple[int, None]]:
        val_arg = received_args[i_token]
        if val_arg == "--":
            return -1, None

        if val_arg.startswith("-"):
            return self._parse_flag(i_token, received_args)
        else:
            return self._parse_positional_argument(val_arg)

    def _parse_flag(
        self,
        i_token: int,
        received_args: List[str],
    ) -> Tuple[int, Dict[str, Arg]]:
        val_arg = received_args[i_token]

        if self.args[val_arg].mode == "count":
            self._parse_count_flag(val_arg)
            return 1, self.args[val_arg]
        if self.args[val_arg].mode == "bool/true/stopper":
            self.args[val_arg].value = True
            return -1, self.args[val_arg]
        if self.args[val_arg].mode == "bool/true":
            self.args[val_arg].value = True
            return 1, self.args[val_arg]
        if self.args[val_arg].mode == "bool/false":
            self.args[val_arg].value = False
            return 1, self.args[val_arg]

        try:
            next_val_arg = received_args[i_token + 1]
        except IndexError as e:
            msg = "Need value for optional arg {val_arg.name}"
            raise RuntimeError(msg) from e

        if self.args[val_arg].mode == "array":
            self._parse_array_flag(val_arg, next_val_arg)
        else:
            self.args[val_arg].value = next_val_arg

        return 2, self.args[val_arg]

    def _parse_array_flag(self, val_arg: Arg, next_val_arg: Arg) -> None:
        if self.args[val_arg].value is None:  # not hasattr(self.args[val_arg], "value")
            self.args[val_arg].value = [next_val_arg]
        else:
            self.args[val_arg].value.append(next_val_arg)

    def _parse_count_flag(self, val_arg: Arg) -> None:
        if self.args[val_arg].value is None:
            self.args[val_arg].value = 1
        else:
            self.args[val_arg].value += 1

    def _parse_positional_argument(self, val_arg: Arg) -> Tuple[int, Dict[str, Arg]]:
        i_pos_name = list(self.args_pos_def.keys())[len(self.args_pos.keys())]
        self.args[i_pos_name].value = val_arg
        self.args_pos[i_pos_name] = self.args[i_pos_name]

        return 1, self.args[i_pos_name]

        # TODO finish check
        # all positional have a value

    def print_help(self) -> None:
        if self.help:
            print(
                Painter.parse_color_tags(f"<ST_UNDERLINE>{self.help}</ST_UNDERLINE>\n"),
            )

        # TODO show command line argument

        info_not_option = []
        info_option = []
        for arg in self.args_def:
            if arg.name.startswith("-"):
                opt = ", ".join([arg.name, *arg.alternate_name])
                next_value = (
                    ""
                    if arg.mode == "count" or arg.mode.startswith("bool/")
                    else ' "value"'
                )
                info_option.append([f"""{opt}{next_value}""", arg.help])
            else:
                info_not_option.append([arg.name, arg.help])

        infos = info_not_option + info_option

        size_col = max_string_length_per_column(infos)

        wrapped = wrap_text_in_table(infos, [size_col[0], 100 - size_col[0]])
        for row in wrapped:
            for ss in [list(sub_row) for sub_row in zip(*row)]:
                print("  " + (" | ".join(ss)).rstrip())
            print()



from uvpipx.internal_libs.colors import Painter

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2024 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.4.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


def show_version():
    print(
        Painter.parse_color_tags(
            f"<BRIGHT_WHITE>uvpipx</BRIGHT_WHITE> version <CYAN>{__version__}</CYAN>\n",
        ),
    )
